#include "nbc_def.h"
#include "nbc_ext.h"

void em_main(void) {
  int u;
  FLOAT_TYPE ic_opt=0.0;
  FILE *fp_param;
  FILE *fp_memp;

  /* Read the occurrence data first */
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  stime_read = current_sec();
  stime_lap_read = elapsed_sec();
  read_freq();
  etime_read = current_sec();
  etime_lap_read = elapsed_sec();
#else
  read_freq();
#endif

  print_size(fp_log); /* Record the model size, etc. */


  /**** Run the EM algorithm: ****/

  if (mode_search_num_class) allocate_ic();
  allocate_seed();

  if (mode_rank_sm) {
    fp_sm = fopen(fname_rank_sm,"wb");
    if (fp_sm == NULL)
      die("Couldn't open %s\n",fname_rank_sm);
  }

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  stime_em = current_sec();
  stime_lap_em = elapsed_sec();
#endif

  /* we attempt to search for the optimal number of classes: */
  if (mode_search_num_class && !mode_read_hparam) {
    u = 0;

    for (num_class = kmin; num_class <= kmax; num_class += kstep) {
      fprintf(fp_msg,"|Classes| = %d:\n",num_class);

      /* (re-)set the smoothing constants according to the number of classes*/
      set_auto_smooth();

      /* allocate memory areas for EM */
      allocate_param();
      allocate_param_best();
      if (!mode_nomem) allocate_prob();
      allocate_expect();
      allocate_log_likelihood();
      allocate_log_likelihood_best();

      /* run the body part of the EM algorithm */
      if (num_sm > 0)
	smem_iterate(u);
      else
	em_iterate(u);

      /* compute the model scores */
      bic[u] = compute_bic();
      if (mode_smooth_strict) autoclass[u] = compute_autoclass();
      switch (ic_type) {
      case BIC:
	ic[u] = bic[u];
	break;
      case AUTOCLASS:
	ic[u] = autoclass[u];
	break;
      default:
	die("Internal error: Unsupported information criterion\n");
      }

      release_param_best();
      if (!mode_nomem) release_prob();
      release_expect();
      release_log_likelihood();
      release_log_likelihood_best();

      /* now we have the (temporarily) optimal number of classes */
      if (num_class == kmin || ic[u] > ic_opt) {
	fprintf(fp_msg,conv_msg("  [%d] %s = %.6Lf (temporarily optimal)\n"),
                num_class,ic_type_str,ic[u]);
	if (num_class > kmin) release_param_opt();
	num_class_opt = num_class;
	ic_opt = ic[u];

	save_param_opt();  /* save the current parameters */

      } else {
	fprintf(fp_msg,conv_msg("  [%d] %s = %.6Lf\n"),num_class,ic_type_str,ic[u]);
      }

      release_param();

      u++;
    }
    fprintf(fp_msg,"Optimal |Classes| = %d\n",num_class_opt);

    /* restore the parameters with the optimal number of classes */
    restore_param_opt();
    release_param_opt();
    num_class = num_class_opt;

    print_ic(fp_log);
    print_seed(fp_log);
  }
  /* we do not try to find the optimal number of classes */
  else {
    if (mode_read_hparam) {
      read_hparam();   /* allocate_hparam() will be called  */
      check_hparam();  /* check if mode_smooth_strict holds */
    }
    else {
      set_auto_smooth();
    }

    fprintf(fp_msg,"#classes = %d:\n",num_class);

    allocate_param();
    allocate_log_likelihood();
    if (!mode_nomem) allocate_prob();
    allocate_expect();
    allocate_param_best();
    allocate_log_likelihood_best();

    if (num_sm > 0) {
      smem_iterate(0);
    } else {
      em_iterate(0);
    }

    if (mode_read_hparam) release_hparam();
    release_param_best();
    if (!mode_nomem) release_prob();
    release_expect();
    release_log_likelihood();
    release_log_likelihood_best();

    print_seed(fp_log);
  }

  if (mode_search_num_class) release_ic();  /* release model scores */
  release_seed();

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  etime_em = current_sec();
  etime_lap_em = elapsed_sec();
#endif

  if (mode_rank_sm) fclose(fp_sm);

  if (mode_verbose) print_param(fp_log);

  /*
   * at this point, num_class, param_c[] and param_w[][] store the values
   * with the optimal number of classes
   */

  /* record the estimated parameters */
  fp_param = fopen(fname_param,"wb");
  if (fp_param == NULL)
    die("Couldn't open %s\n", fname_param);
  print_param(fp_param);
  fclose(fp_param);

  /* compute and record the membership probabilities */
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  if (mode_memp) stime_memp = current_sec();
  if (mode_memp) stime_lap_memp = elapsed_sec();
#endif

  if (mode_memp || mode_cluster || mode_rank || mode_dist) {
    allocate_prob_mem();
    compute_member_prob();
  }
  if (mode_memp) {
    fp_memp = fopen(fname_memp,"wb");
    if (fp_memp == NULL)
      die("Couldn't open %s\n", fname_memp);
    print_member_prob(fp_memp);
    fclose(fp_memp);
  }

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  if (mode_memp) etime_memp = current_sec();
  if (mode_memp) etime_lap_memp = elapsed_sec();
#endif

  if (mode_dist) dist_main();        /* computing dissimilarities */
  if (mode_rank) rank_main();        /* relevance analysis */
  if (mode_cluster) cluster_main();  /* clustering */

  release_param();
  if (mode_memp || mode_cluster || mode_rank || mode_dist)
    release_prob_mem();

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  print_time(fp_log); /* record the computation time */
#endif
}

/*
 * we do not run the EM algorithm, but read the parameters obtained
 * by a previous run of nbc
 */
void no_em_main(void)
{
  FILE *fp_memp;

  /** read the parameters (allocate_param() is called) **/
  fprintf(fp_msg,"Reading parameters from %s...", fname_param);
  fflush(fp_msg);
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  stime_read = current_sec();
  stime_lap_read = elapsed_sec();
  read_param();
  etime_read = current_sec();
  etime_lap_read = elapsed_sec();
#else
  read_param();
#endif
  fprintf(fp_msg,"done\n");

  if (mode_cluster) {
    fprintf(fp_msg,"Reading cases from %s...", fname_freq);
    fflush(fp_msg);
    read_freq();
    fprintf(fp_msg,"done\n");
  }

  set_auto_smooth();

  print_size(fp_log);

  /* computing membership probabilities */
  if (mode_memp) {
    fprintf(fp_msg,"Computing the membership distribution...");
    fflush(fp_msg);
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
    stime_memp = current_sec();
    stime_lap_memp = elapsed_sec();
#endif
  }
  if (mode_memp || mode_cluster || mode_rank || mode_dist) {
    allocate_prob_mem();
    compute_member_prob();
  }
  if (mode_memp) {
    fp_memp = fopen(fname_memp,"wb");
    if (fp_memp == NULL)
      die("Couldn't open %s\n", fname_memp);
    print_member_prob(fp_memp);
    fclose(fp_memp);
  }
  
  if (mode_memp) {
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
    etime_memp = current_sec();
    etime_lap_memp = elapsed_sec();
#endif
    fprintf(fp_msg,"done.\n");
  }

  /* computing dissimilarities */
  if (mode_dist) {
    fprintf(fp_msg,"Computing the (dis)similarities between attribute values...");
    fflush(fp_msg);
    dist_main();
    fprintf(fp_msg,"done.\n");
  }

  /* relevance analysis */
  if (mode_rank) {
    fprintf(fp_msg,"Ranking relevant words...");
    fflush(fp_msg);
    rank_main();
    fprintf(fp_msg,"done.\n");
  }

  /* clustering */
  if (mode_cluster) {
    fprintf(fp_msg,"Clustering cases and attribute values...");
    fflush(fp_msg);
    cluster_main();
    fprintf(fp_msg,"done.\n");
  }

  /* cleaning-up */
  release_param();
  if (mode_memp || mode_cluster || mode_rank || mode_dist)
    release_prob_mem();

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  print_time(fp_log);
#endif
}

/*
 * EM run(s) with some paticular number of classes
 * -- u is the index for searching the optimal number of classes
 */
void em_iterate(int u)
{
  int r,r_best=-1,n;
  int num_iterate,num_iterate_best=-1;
  unsigned int curr_seed=0;

  for (r=0; r<num_init; r++) { /* random restarts */
    if (!mode_set_seed) {
      curr_seed = (unsigned int)rand(); /* reset the random seeds */
      srand(curr_seed);
    }

    /* initialization step */
    initialize_param();
    
    /* get the max number of EM iterations */
    n = (num_init > 1) ? max_init_iterate : max_iterate;

    /* run the body of EM, and get the number of iterations */
    if (mode_nomem) {
      num_iterate = em_iterate_body_nomem(r,0,n);
    }
    else {
      num_iterate = em_iterate_body(r,0,n);
    }

    /* record the results obtained by a random restart */
    log_seed[u][r] = (!mode_set_seed) ? curr_seed : 0;
    log_num_iterate[u][r] = num_iterate;
    log_final_lambda[u][r] = final_lambda;

    /* save the current estimate if we have the highest likelihood (a positeriori prob) */
    if (r == 0 || final_lambda > final_lambda_best) {
      r_best = r;
      final_lambda_best = final_lambda;
      final_log_likelihood_best = final_log_likelihood;
      final_log_prior_best = final_log_prior;
      num_iterate_best = num_iterate;
      save_log_likelihood_best(num_iterate_best);
      save_param_best();
    }
  }

  /* record the index of the current restart */
  log_r_best[u] = r_best;

  /* restore the parameters with the highest likelihood (a posteriori prob), etc. */
  final_lambda = final_lambda_best;
  final_log_likelihood = final_log_likelihood_best;
  final_log_prior = final_log_prior_best;
  restore_param_best();
  restore_log_likelihood_best(num_iterate_best);
  num_iterate = num_iterate_best;

  /* continue the EM iterations */
  if (num_init > 1) {
    fprintf(fp_msg,"  <<Resumed best parameter set #%d>>\n",r_best);
    if (mode_nomem) {
      num_iterate = em_iterate_body_nomem(r_best,num_iterate,max_iterate);
    } else {
      num_iterate = em_iterate_body(r_best,num_iterate,max_iterate);
    }
  }

  /* output the likelihood (the a posteriori prob) */
  print_log_likelihood(fp_log,num_class,num_iterate);
}

/* 
 * body of the EM algorithm
 * - r: index of restarts
 * - m: starting index of EM iterations
 * - n: max number of iterations
 */
int em_iterate_body(int r, int m, int n)
{
  compute_complete_prob();
  compute_incomplete_prob();

  log_likelihood[m] = compute_log_likelihood();

  /* lambda = log of a posteriori prob (MAP) or log-likelihood (ML) */
  if (mode_smooth) {
    log_prior[m] = compute_log_prior();
    lambda[m] = log_likelihood[m] + log_prior[m];
  }
  else {
    lambda[m] = log_likelihood[m];
  }
  
  if (num_init > 1) {
    fprintf(fp_msg,"  [%d] #iterations ",r);
  }
  else {
    fprintf(fp_msg,"  #iterations ");
  }
  print_dot_message(m);
    
  /* Iterate E-step and M-step */
  while (1) {
    compute_conditional_prob();  /* compute the conditional probs */
    compute_expect();            /* compute the expected occurrences */
    update_param();              /* update the parameters (M-step) */
    
    m++;
   
    compute_complete_prob();     /* compute the complete probs under the updated params */
    compute_incomplete_prob();   /* compute the incomplete probs under the updated params */
   
    log_likelihood[m] = compute_log_likelihood();
    
    if (mode_smooth) {
      log_prior[m] = compute_log_prior();
      lambda[m] = log_likelihood[m] + log_prior[m];
    } else {
      lambda[m] = log_likelihood[m];
    }
    
    print_dot_message(m);

    if (is_decreased(r,m,n)) break;
    if (is_converged(r,m,n)) break;
  }
  
  /* record the converged likelihood (a posteriori prob) */
  final_log_likelihood = log_likelihood[m];
  final_lambda = lambda[m];
  if (mode_smooth) final_log_prior = log_prior[m];

  /* return the index of the next iteration
   * (not the number of iterations conducted in this function)
   */
  return m;
}

/*
 * This function produces the same result as that of em_iterate_body(),
 * but runs with less memory than em_iterate_body().
 */
int em_iterate_body_nomem(int r, int m, int n)
{
  if (num_init > 1) {
    fprintf(fp_msg,"  [%d] #iterations ",r);
  } else {
    fprintf(fp_msg,"  #iterations ");
  }
  print_dot_message(m);

  log_likelihood[m] = compute_expect_nomem();

  if (mode_smooth) {
    log_prior[m] = compute_log_prior();
    lambda[m] = log_likelihood[m] + log_prior[m];
  } else {
    lambda[m] = log_likelihood[m];
  }

  while (1) {
    update_param();
    m++;
    print_dot_message(m);
    log_likelihood[m] = compute_expect_nomem();
    
    if (mode_smooth) {
      log_prior[m] = compute_log_prior();
      lambda[m] = log_likelihood[m] + log_prior[m];
    } else {
      lambda[m] = log_likelihood[m];
    }
    
    if (is_decreased(r,m,n)) break;
    if (is_converged(r,m,n)) break;
  }
  
  final_log_likelihood = log_likelihood[m];
  final_lambda = lambda[m];
  if (mode_smooth) final_log_prior = log_prior[m];

  return m;
}

int is_decreased(int r, int m, int n)
{
  /* This would not happen without a bug or a numeric error */
  if (lambda[m] < lambda[m-1] && (lambda[m-1] - lambda[m]) > ACCEPT_SMALL_NUMBER) {
    print_decrease(fp_log,r,m,lambda[m-1],lambda[m]);
    fprintf(fp_msg,"\nLL or LPost decreased! -- see the log file\n");
    return 1;
  }
  return 0;
}

int is_converged(int r, int m, int n)
{
  if (m >= MAX_MAX_ITERATE-1) {
    fprintf(fp_msg,conv_msg(" (Stopped: %d iterations)\tL=%.6Lf\n"),m,lambda[m]);
    return 1;
  } else if (n > 0 && m >= n) {
    fprintf(fp_msg,conv_msg(" (Stopped: %d iterations)\tL=%.6Lf\n"),m,lambda[m]);
    return 1;
  } else if (max_iterate > 0 && m >= max_iterate) {
    fprintf(fp_msg,conv_msg(" (Stopped: %d iterations)\tL=%.6Lf\n"),m,lambda[m]);
    return 1;
  } else {
    if (lambda[m] - lambda[m-1] < epsilon) {
      fprintf(fp_msg,conv_msg(" (Converged: %d iterations)\tL=%.6Lf\n"),m,lambda[m]);
      return 1;
    }
  }
  return 0;
}

void compute_incomplete_prob(void)
{
  int k,e;

#pragma omp parallel for private(k)
  for (e=0; e<num_entry; e++) {
    prob_incomp[e] = 0.0;
    for (k=0; k<num_class; k++)
      prob_incomp[e] += prob_comp[k][e];
  }

  if (mode_verbose) print_incomplete_prob(fp_log);
}

FLOAT_TYPE compute_log_likelihood(void)
{
  FLOAT_TYPE ll;
  int e, f;

  ll = 0.0;
  /* doesn't work if a `pragma omp' statement is annotated (i don't know why) */
  for (e=0; e<num_entry; e++) {
    f = freq_list[e].freq;
    if (f > 0) ll += f * MATH_LOG(prob_incomp[e]);
  }

  return ll;
}

void compute_conditional_prob(void)
{
  int k,e;
  FLOAT_TYPE p;

#pragma omp parallel for private(e,p)
  for (k=0; k<num_class; k++) {
    for (e=0; e<num_entry; e++) {
      p = prob_incomp[e];
      prob_cond[k][e] = (p > 0) ? prob_comp[k][e]/p : 0.0;
    }
  }

  if (mode_verbose) print_conditional_prob(fp_log);
}

void save_log_likelihood_best(int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++) {
    log_likelihood_best[m] = log_likelihood[m];
    lambda_best[m] = lambda[m];
  }

  if (mode_smooth)
    for (m=0; m<num_iterate; m++) {
      log_prior_best[m] = log_prior[m];
    }
}

void restore_log_likelihood_best(int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++) {
    log_likelihood[m] = log_likelihood_best[m];
    lambda[m] = lambda_best[m];
  }

  if (mode_smooth)
    for (m=0; m<num_iterate; m++) {
      log_prior[m] = log_prior_best[m];
    }
}

void initialize_param(void)
{
  int k,v,i;
  FLOAT_TYPE r;
  FLOAT_TYPE sum;
  FLOAT_TYPE noise_c,noise_w;

  if (init_type_class == RANDOM_C) {
    sum = 0.0;
    for (k=0; k<num_class; k++) {
      r = (rand() % RAND_MAX) + RAND_MAX;
      param_c[k] = r;
      sum += r;
    }
    for (k=0; k<num_class; k++) {
      param_c[k] /= sum;
    }
  }
  else if (init_type_class == NOISY_UNIFORM_C) {
    sum = 0.0;
    for (k=0; k<num_class; k++) {
      noise_c = NOISE_RATIO*(rand() % RAND_MAX)/(FLOAT_TYPE)RAND_MAX;
      r = (1.0/num_class)*(1+noise_c);
      param_c[k] = r;
      sum += r;
    }
    for (k=0; k<num_class; k++) {
      param_c[k] /= sum;
    }
  }

  if (init_type_word == RANDOM_W) {
    for (v=0; v<tuple_size; v++) {
      for (k=0; k<num_class; k++) {
        sum = 0.0;
        for (i=(num_word[v]-1); i>=0; i--) {
          r = (rand() % RAND_MAX) + RAND_MAX;
          param_w[k][v][i] = r;
          sum += r;
        }
        for (i=(num_word[v]-1); i>=0; i--) {
          param_w[k][v][i] /= sum;
        }
      }
    }
  }
  else if (init_type_word == NOISY_UNIFORM_W) {
    for (v=0; v<tuple_size; v++) {
      for (k=0; k<num_class; k++) {
        sum = 0.0;
        for (i=(num_word[v]-1); i>=0; i--) {
          noise_w = NOISE_RATIO*(rand() % RAND_MAX)/(FLOAT_TYPE)RAND_MAX;
          r = (1.0 / num_word[v])*(1 + noise_w);
          param_w[k][v][i] = r;
          sum += r;
        }
        for (i=(num_word[v]-1); i>=0; i--) {
          param_w[k][v][i] /= sum;
        }
      }
    }
  }
  else if (init_type_word == WORD_FREQ) {
    for (v=0; v<tuple_size; v++) {
      for (k=0; k<num_class; k++) {
        sum = 0.0;
        for (i=(num_word[v]-1); i>=0; i--) {
          noise_w = NOISE_RATIO*(rand() % RAND_MAX)/(FLOAT_TYPE)RAND_MAX;
          r = word_freq_list[v][i] * (1 + noise_w);
          param_w[k][v][i] = r;
          sum += r;
        }
        for (i=(num_word[v]-1); i>=0; i--) {
          param_w[k][v][i] /= sum;
        }
      }
    }
  }
  else {
    fprintf(fp_msg,"Internal error: unexpected initialization method for word\n");
  }
  
  if (mode_verbose) print_param(fp_log);
}

void compute_complete_prob(void)
{
  int k,e,v;

#pragma omp parallel for private(e,v)
  for (k=0; k<num_class; k++) {
    for (e=0; e<num_entry; e++) {
      prob_comp[k][e] = param_c[k];
      for (v=0; v<tuple_size; v++) {
	prob_comp[k][e] *= param_w[k][v][freq_list[e].id[v]];
      }
    }
  }

  if (mode_verbose) print_complete_prob(fp_log);
}

/* memory-less version */
FLOAT_TYPE compute_log_likelihood_nomem(void)
{
  FLOAT_TYPE ll;
  int k,e,f,v;
  FLOAT_TYPE p_ic,p_c;

  ll = 0.0;
  /* Cannot immediately add a `pragma omp' statement here */
  for (e=0; e<num_entry; e++) {
    f = freq_list[e].freq;
    if (f <= 0) continue;

    p_ic = 0.0;
    for (k=0; k<num_class; k++) {
      p_c = param_c[k];
      for (v=0; v<tuple_size; v++) {
	p_c *= param_w[k][v][freq_list[e].id[v]];
      }
      p_ic += p_c;
    }

    ll += f * MATH_LOG(p_ic);
  }

  return ll;
}

FLOAT_TYPE compute_log_prior(void)
{
  FLOAT_TYPE lp,lp_c,lp_w;
  int k,i,v;
  FLOAT_TYPE alpha_sum;

  lp = 0.0;

  if (mode_read_hparam) {
    lp_c = 0.0;

    /* Compute the normalizing constant for class */
    alpha_sum = 0.0;
    for (k=0; k<num_class; k++) {
      alpha_sum += hparam_c[k];
    }
    lp_c += LGAMMA(alpha_sum);
    for (k=0; k<num_class; k++) {
      lp_c -= LGAMMA(hparam_c[k]);
    }

    for (k=0; k<num_class; k++)
      lp_c += (hparam_c[k] - 1.0) * MATH_LOG(param_c[k]);

    lp += lp_c;
  }
  else if (smooth_c > 0) {
    lp_c = 0.0;

    lp_c += LGAMMA(num_class * (smooth_c + 1.0));
    lp_c -= num_class * LGAMMA(smooth_c + 1.0);

    for (k=0; k<num_class; k++)
      lp_c += smooth_c * MATH_LOG(param_c[k]);

    lp += lp_c;
  }

  if (mode_read_hparam) {
    lp_w = 0.0;

    /* Compute the normalizing constant */
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size; v++) {
	alpha_sum = 0.0;
	for (i=0; i<num_word[v]; i++) {
	  alpha_sum += hparam_w[k][v][i];
	}
	lp_w += LGAMMA(alpha_sum);
      }
    }
    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size; v++)
	for (i=0; i<num_word[v]; i++)
	  lp_w -= LGAMMA(hparam_w[k][v][i]);

    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size; v++)
	for (i=0; i<num_word[v]; i++)
	  lp_w += (hparam_w[k][v][i] - 1.0) * MATH_LOG(param_w[k][v][i]);

    lp += lp_w;
  }
  else {
    lp_w = 0.0;

    /* Compute the normalizing constant */
    for (v=0; v<tuple_size; v++) {
      for (k=0; k<num_class; k++) {
	lp_w += LGAMMA(num_word[v] * (smooth_w[v] + 1.0));
	lp_w -= num_word[v] * LGAMMA(smooth_w[v] + 1.0);
      }
    }

    for (v=0; v<tuple_size; v++) {
      if (smooth_w[v] > 0) {
	for (k=0; k<num_class; k++)
	  for (i=0; i<num_word[v]; i++)
	    lp_w += smooth_w[v] * MATH_LOG(param_w[k][v][i]);
      }
    }

    lp += lp_w;
  }

  return lp;
}

/* memory-less version of compute_expect() */
FLOAT_TYPE compute_expect_nomem(void)
{
  FLOAT_TYPE ll;
  int k,e,f,i,v;
  FLOAT_TYPE p_ic,p_c,p_e;

  #pragma omp parallel for private(v,i)
  for (k=0; k<num_class; k++) {
    expect_c[k] = 0.0;
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
	expect_w[k][v][i] = 0.0;
      }
    }
  }

  ll = 0.0;

  /* Cannot immediately add a `pragma omp' statement */
  /* #pragma omp parallel for private(k,v,p_ic,f,p_e,p_c) reduction(+: ll) */
  for (e=0; e<num_entry; e++) {
    p_ic = 0.0;
    f = freq_list[e].freq;
    
    for (v=0; v<tuple_size; v++) {
      freq_w_id[v] = freq_list[e].id[v];
    }

    for (k=0; k<num_class; k++) {
      p_c = param_c[k];
      for (v=0; v<tuple_size; v++)
	p_c *= param_w[k][v][freq_w_id[v]];
      p_ic += p_c;
    }    

    for (k=0; k<num_class; k++) {
      p_e = f * param_c[k];
      for (v=0; v<tuple_size; v++)
	p_e *= param_w[k][v][freq_w_id[v]];
      p_e /= p_ic;
      expect_c[k] += p_e;
      for (v=0; v<tuple_size; v++)
	expect_w[k][v][freq_w_id[v]] += p_e;
    }

    ll += f * MATH_LOG(p_ic);
  }
  
  if (mode_verbose) print_expectation(fp_log);

  return ll;
}

void compute_expect(void)
{
  int k,i,v,e;
  FLOAT_TYPE tmp_e;

#pragma omp parallel for private(e)
  for (k=0; k<num_class; k++) {
    expect_c[k] = 0.0;
    for (e=0; e<num_entry; e++)
      expect_c[k] += freq_list[e].freq * prob_cond[k][e];
  }

#pragma omp parallel for private(i,e,v,tmp_e)
  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++)
      for (i=0; i<num_word[v]; i++)
        expect_w[k][v][i] = 0.0;

    for (e=0; e<num_entry; e++) {
      tmp_e = freq_list[e].freq * prob_cond[k][e];
      for (v=0; v<tuple_size; v++)
        expect_w[k][v][freq_list[e].id[v]] += tmp_e;
    }
  }

  if (mode_verbose) print_expectation(fp_log);
}

void update_param(void)
{
  FLOAT_TYPE denom_c;
  FLOAT_TYPE denom_w;
  int k,v,i;

  if (mode_read_hparam) {
    denom_c = 0.0;
    for (k=0; k<num_class; k++)
      denom_c += (param_c[k] = expect_c[k] + hparam_c[k] - 1.0);
    for (k=0; k<num_class; k++)
      param_c[k] /= denom_c;
  }
  else if (smooth_c > 0) {
    denom_c = total_freq + smooth_c * num_class;
#pragma omp parallel for
    for (k=0; k<num_class; k++)
      param_c[k] = (expect_c[k] + smooth_c)/denom_c;
  }
  else {
#pragma omp parallel for
    for (k=0; k<num_class; k++)
      param_c[k] = expect_c[k]/total_freq;
  }

  if (mode_read_hparam) {
#pragma omp parallel for private(i,v,denom_w)
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size; v++) {
	denom_w = 0.0;
	for (i=0; i<num_word[v]; i++)
	  denom_w += (param_w[k][v][i] = expect_w[k][v][i] + hparam_w[k][v][i] - 1.0);
	for (i=0; i<num_word[v]; i++)
	  param_w[k][v][i] /= denom_w;
      }
    }
  }
  else {
    for (v=0; v<tuple_size; v++) {
      if (smooth_w[v] > 0) {
#pragma omp parallel for private(i,denom_w)
	for (k=0; k<num_class; k++) {
	  denom_w = expect_c[k] + smooth_w[v] * num_word[v];
	  for (i=0; i<num_word[v]; i++)
	    param_w[k][v][i] = (expect_w[k][v][i] + smooth_w[v])/denom_w;
	}
      }
      else {
#pragma omp parallel for private(i,denom_w)
	for (k=0; k<num_class; k++) {
	  denom_w = expect_c[k];
	  for (i=0; i<num_word[v]; i++)
	    param_w[k][v][i] = expect_w[k][v][i]/denom_w;
	}
      }
    }
  }

  if (mode_verbose) print_param(fp_log);
}

void check_param(void)
{
  int k,v,i;
  FLOAT_TYPE sum_c, sum_w;

  sum_c = 0.0;
  for (k=0; k<num_class; k++) {
    sum_c += param_c[k];
  }
  fprintf(fp_msg,conv_msg("check> sum_c = %.9Lf\n"),sum_c);

  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      sum_w = 0.0;
      for (i=0; i<num_word[v]; i++) {
        sum_w += param_w[k][v][i];
      }
      fprintf(fp_msg,conv_msg("check> sum_w1[%d] = %.9Lf\n"),k,sum_w);
    }
  }
}

void compute_member_prob(void)
{
  int k,v,i,e;
  FLOAT_TYPE p_c, p_ic;

#pragma omp parallel for private(i)
  for (v=0; v<tuple_size; v++) {
    for (i=0; i<num_word[v]; i++)
      prob_w[v][i] = 0.0;
  }

  /* Cannot immediately add a `pragma omp' statement here */
  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
        prob_w[v][i] += (prob_att_mem[k][v][i] = param_c[k] * param_w[k][v][i]);
      }
    }
  }

  /* normalization */
#pragma omp parallel for private(v,i)
  for (k=0; k<num_class; k++)
    for (v=0; v<tuple_size; v++)
      for (i=0; i<num_word[v]; i++)
        prob_att_mem[k][v][i] /= prob_w[v][i];

#pragma omp parallel for private(v,p_ic,p_c,k)
  for (e=0; e<num_entry; e++) {
    for (v=0; v<tuple_size; v++)
      freq_w_id[v] = freq_list[e].id[v];

    p_ic = 0.0;
    for (k=0; k<num_class; k++) {
      p_c = param_c[k];
      for (v=0; v<tuple_size; v++)
	p_c *= param_w[k][v][freq_w_id[v]];
      prob_case_mem[k][e] = p_c;
      p_ic += p_c;
    }
    
    for (k=0; k<num_class; k++) {
      prob_case_mem[k][e] /= p_ic;
    }
  }
}

/*
 * Since the number of parameters can vary, we should allocate
 * the new memory area (note that the size is based on num_class_opt).
 */
void save_param_opt(void)
{
  int k,v,i;

  param_c_opt = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class_opt);
  param_w_opt = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class_opt);
  if (param_c_opt == NULL || param_w_opt == NULL)
    die("Couldn't allocate memory for param_c_opt or param_w_opt\n");

#pragma omp parallel for private(v)
  for (k=0; k<num_class_opt; k++) {
    param_w_opt[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    if (param_w_opt[k] == NULL)
      die("Couldn't allocate memory for param_w_opt[k]\n");
    for (v=0; v<tuple_size; v++) {
      param_w_opt[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      if (param_w_opt[k][v] == NULL)
        die("Couldn't allocate memory for param_w_opt[k][v]\n");
    }
  }

#pragma omp parallel for private(v,i)
  for (k=0; k<num_class_opt; k++) {
    param_c_opt[k] = param_c[k];
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
	param_w_opt[k][v][i] = param_w[k][v][i];
      }
    }
  }
}

/*
 * Since the number of parameters can vary, we should allocate
 * the new memory area (note that the size is based on num_class_opt).
 */
void restore_param_opt(void)
{
  int k,v,i;

  param_c = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class_opt);
  param_w = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class_opt);
  if (param_c == NULL || param_w == NULL)
    die("Couldn't allocate memory for param_c or param_w\n");

#pragma omp parallel for private(v)
  for (k=0; k<num_class_opt; k++) {
    param_w[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    if (param_w[k] == NULL)
      die("Couldn't allocate memory for param_w[k]\n");
    for (v=0; v<tuple_size; v++) {
      param_w[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      if (param_w[k][v] == NULL)
        die("Couldn't allocate memory for param_w[k][v]\n");
    }
  }

#pragma omp parallel for private(i)
  for (k=0; k<num_class_opt; k++) {
    param_c[k] = param_c_opt[k];
    for (v=0; v<tuple_size; v++)
      for (i=0; i<num_word[v]; i++)
        param_w[k][v][i] = param_w_opt[k][v][i];
  }
}

void save_param_best(void)
{
  int k,v,j;

#pragma omp parallel for private(v,j)
  for (k=0; k<num_class; k++) {
    param_c_best[k] = param_c[k];
    for (v=0; v<tuple_size; v++) {
      for (j=0; j<num_word[v]; j++) {
	param_w_best[k][v][j] = param_w[k][v][j];
      }
    }
  }
}

void restore_param_best(void)
{
  int k,v,j;

#pragma omp parallel for private(v,j)
  for (k=0; k<num_class; k++) {
    param_c[k] = param_c_best[k];
    for (v=0; v<tuple_size; v++) {
      for (j=0; j<num_word[v]; j++) {
	param_w[k][v][j] = param_w_best[k][v][j];
      }
    }
  }
}

void check_hparam(void)
{
  int k,v,i;

  mode_smooth = 0;
  mode_smooth_strict = 1;
  for (k=0; k<num_class; k++) {
    mode_smooth |= (hparam_c[k] > 1.0) ? 1 : 0;
    mode_smooth_strict &= (hparam_c[k] > 1.0) ? 1 : 0;
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
	mode_smooth |= (hparam_w[k][v][i] > 1.0) ? 1 : 0;
	mode_smooth_strict &= (hparam_w[k][v][i] > 1.0) ? 1 : 0;
      }
    }
  }

  if (!mode_smooth_strict) {
    fprintf(stderr,"Pseudo counts read from %s should be all positive\n",fname_smooth);
    print_usage(stderr);
    exit(1);
  }
}

