#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#ifdef USE_ICONV
#  include <iconv.h>
#endif

#include "getopt/getopt.h"

#ifdef USE_FREE_GETOPT
#  ifdef USE_ICONV
#    define OPT_STRING  "-f:s:e:p:q:o:v:m:lh"
#  else
#    define OPT_STRING  "-f:s:e:o:v:m:lh"
#  endif
#else
#  ifdef USE_ICONV
#    define OPT_STRING  "f:s:e:p:q:o:v:m:lh"
#  else
#    define OPT_STRING  "f:s:e:o:v:m:lh"
#  endif
#endif
#define MAX_IN_FILENAME  1024
#define MAX_EXTENSION    10
#define MAX_ID           1024
#define DEFAULT_BASE     "out"
#define DEFAULT_EXT      "out"
#define CHARCODE_LEN      100
#define MAX_LINE_LEN     65536
#define MAX_OUT_LEN      65536
#ifdef USE_ICONV
#  define USAGE "[Usage] nbcsep [-f BASE][-s SUFFIX][-e EXT]" \
                "[-v ID_TO_SHOW][-o ID_TO_SAVE][-m MATCH_TYPE]" \
                "[-p IN_CHARCODE][-q OUT_CHARCODE] [-h] DATA\n" 
#  define CHARCODE_INTERNAL "UTF-8"
#else
#  define USAGE "[Usage] nbcsep [-f BASE][-s SUFFIX][-e EXT]" \
                "[-v ID_TO_SHOW][-o ID_TO_SAVE][-m MATCH_TYPE] [-h] DATA\n"
#endif

enum match_type {MATCH_INFIX, MATCH_EXACT, MATCH_PREFIX, MATCH_SUFFIX};

char fname_input[MAX_IN_FILENAME];
char fname_base[MAX_IN_FILENAME];
char fname_suffix[MAX_IN_FILENAME];
char fname_ext[MAX_EXTENSION];
char id_str[MAX_ID];
int mode_ext;
int mode_suffix;
int mode_id;
int mode_list;
int mode_stdin;
int mode_stdout;
int mode_conv_in;
int mode_conv_out;
enum match_type mode_match;
#ifdef USE_ICONV
iconv_t cd_in, cd_out;
char charcode_in[CHARCODE_LEN];
char charcode_out[CHARCODE_LEN];
#endif
char *buf_mid;
char *buf_out;

void print_usage(FILE *fp)
{
  fprintf(fp,USAGE);
}

void initialize_modes(void)
{
  strcpy(fname_base,DEFAULT_BASE);
  mode_ext = 0;
  mode_suffix = 0;
  mode_id = 0;
  mode_list = 1;
  mode_stdin = 0;
  mode_stdout = 0;
  mode_match = MATCH_INFIX;
#ifdef USE_ICONV
#ifdef CHARCODE
  strncpy(charcode_in,CHARCODE,CHARCODE_LEN);
  strncpy(charcode_out,CHARCODE,CHARCODE_LEN);
#else
  strncpy(charcode_in,CHARCODE_INTERNAL,CHARCODE_LEN);
  strncpy(charcode_out,CHARCODE_INTERNAL,CHARCODE_LEN);
#endif
#endif
}

void initialize_buffer(void)
{
  buf_mid = (char *)malloc(sizeof(char)*MAX_OUT_LEN);
  if (buf_mid == NULL) {
    fprintf(stderr,"Could't allocate memory for buf_mid\n");
    exit(1);
  }
  buf_out = (char *)malloc(sizeof(char)*MAX_OUT_LEN);
  if (buf_out == NULL) {
    fprintf(stderr,"Could't allocate memory for buf_out\n");
    exit(1);
  }
}

void release_buffer(void)
{
  free(buf_mid); buf_mid = NULL;
  free(buf_out); buf_out = NULL;
}

#ifdef USE_ICONV
void initialize_iconv(void)
{
  cd_in = iconv_open(CHARCODE_INTERNAL,charcode_in);
  if (cd_in == (iconv_t)-1) {
    fprintf(stderr,
	    "iconv_open failed (in: %s)"
	    " -- run 'iconv -l' for available character codes\n",
	    charcode_in);
    exit(1);
  }

  cd_out = iconv_open(charcode_out,CHARCODE_INTERNAL);
  if (cd_out == (iconv_t)-1) {
    fprintf(stderr,
	    "iconv_open failed (out: %s)"
	    " -- run 'iconv -l' for available character codes\n",
	    charcode_out);
    exit(1);
  }
}
#endif

void get_file_extension(char *fname)
{
  int i,len,has_dot;
  char *ext = NULL;

  if (mode_stdin) {
    strcpy(fname_ext,DEFAULT_EXT);
    return;
  }

  len = strlen(fname);
  has_dot = 0;

  for (i=0; i<len; i++) {
    if (fname[i] == '.') {
      has_dot = 1;
      if (i < (len - 1)) {
	i++;
	ext = fname + i;
      }
      else {
	ext = "";
      }
    }
  }

  if (has_dot)
    strcpy(fname_ext,ext);
  else
    strcpy(fname_ext,DEFAULT_EXT);
}

void parse_args(int argc, char *argv[])
{
  int c;
  int i;
#ifdef USE_FREE_GETOPT
  int has_input = 0;
#endif

  while (1) {
    c = getopt(argc,argv,OPT_STRING);
    if (c == -1) break;

    switch (c) {
    case 'f':
      strncpy(fname_base,optarg,(MAX_IN_FILENAME - 1));
      mode_list = 0;
      break;
    case 's':
      mode_suffix = 1;
      strncpy(fname_suffix,optarg,(MAX_IN_FILENAME - 1));
      break;
    case 'e':
      mode_ext = 1;
      strncpy(fname_ext,optarg,(MAX_EXTENSION - 1));
      break;
    case 'o':
      mode_id = 1;
      mode_list = 0;
      strncpy(id_str,optarg,(MAX_ID - 1));
      break;
    case 'v':
      mode_stdout = 1;
      mode_id = 1;
      mode_list = 0;
      strncpy(id_str,optarg,(MAX_ID - 1));
      break;
    case 'm':
      if (strcmp(optarg,"infix") == 0 || strcmp(optarg,"INFIX") == 0) {
	mode_match = MATCH_INFIX;
      }
      else if (strcmp(optarg,"exact") == 0 || strcmp(optarg,"EXACT") == 0) {
	mode_match = MATCH_EXACT;
      }
      else if (strcmp(optarg,"prefix") == 0 || strcmp(optarg,"PREFIX") == 0) {
	mode_match = MATCH_PREFIX;
      }
      else if (strcmp(optarg,"suffix") == 0 || strcmp(optarg,"SUFFIX") == 0) {
	mode_match = MATCH_SUFFIX;
      }
      else {
	fprintf(stderr,"Unknown matching criterion -- %s\n",optarg);
	print_usage(stderr);
	exit(1);
      }
      break;
#ifdef USE_ICONV
    case 'p':
      strncpy(charcode_in,optarg,(CHARCODE_LEN - 1));
      break;
    case 'q':
      strncpy(charcode_out,optarg,(CHARCODE_LEN - 1));
      break;
#endif
    case 'h':
      print_usage(stdout);
      exit(0);
      break;
#ifdef USE_FREE_GETOPT
    case 1:
      if (strncmp(optarg,"-",(MAX_IN_FILENAME - 1)) == 0) {
	mode_stdin = 1;
      }
      else {
	mode_stdin = 0;
	strncpy(fname_input,optarg,(MAX_IN_FILENAME - 1));
      }
      has_input = 1;
      break;
#endif
    default:
      fprintf(stderr,"Unknown option(s) specified\n");
      print_usage(stderr);
      exit(1);
    }
  }

#ifdef USE_FREE_GETOPT
  if (!has_input) {
    fprintf(stderr,"No input file specified\n");
    print_usage(stderr);
    exit(1);
  }
#else
  argc -= optind;
  argv += optind;
  if (argc < 1) {
    fprintf(stderr,"No input file specified\n");
    print_usage(stderr);
    exit(1);
  }
  if (strncmp(argv[0],"-",(MAX_IN_FILENAME - 1)) == 0) {
    mode_stdin = 1;
  }
  else {
    mode_stdin = 0;
    strncpy(fname_input,argv[0],(MAX_IN_FILENAME - 1));
  }
#endif

#ifdef USE_ICONV
  mode_conv_in = (strcmp(charcode_in,CHARCODE_INTERNAL) == 0) ? 0 : 1;
  mode_conv_out = (strcmp(charcode_out,CHARCODE_INTERNAL) == 0) ? 0 : 1;
#else
  mode_conv_in = 0;
  mode_conv_out = 0;
#endif

  if (!mode_ext)
    get_file_extension(fname_input);

  /* Convert the specified ID into the upper case */
  for (i = 0; i < strlen(id_str); i++) {
    id_str[i] = toupper(id_str[i]);
  }
}

char *get_comment_id(char *buf)
{
  char *comment_id;
  char *b;
  char *q;
  int l;

  comment_id = (char *)malloc(sizeof(char)*(MAX_ID + 1));
  if (comment_id == NULL) {
    fprintf(stderr,"Couldn't allocate memory for comment_id\n");
    exit(1);
  }

  b = buf;
  
  if (*b == '#') b++; else { free(comment_id); return NULL; }
  while (*b == '#') b++;
  
  if (*b == ' ') b++; else { free(comment_id); return NULL; }
  while (*b == ' ') b++;
  
  if (*b == 'I') b++; else { free(comment_id); return NULL; }
  if (*b == 'D') b++; else { free(comment_id); return NULL; }
  if (*b == ':') b++; else { free(comment_id); return NULL; }
  
  q = comment_id;
  l = 0;
  while (*b != '\0' && *b != ' ' && *b != '\t' &&
	 *b != '\r' && *b != '\n') {
    if (l >= MAX_ID) {
      fprintf(stderr,"Internal error: too long data ID\n");
      exit(1);
    }
    *q = *b;
    q++;
    b++;
  }
  *q = '\0';
    
  return comment_id;
}

char *build_out_filename(char *comment_id)
{
  char *fname_out;
  int name_len;
  int c;

  name_len = strlen(fname_base);
  if (mode_suffix) name_len += strlen(fname_suffix) + 1;
  name_len += strlen(comment_id) + 2;
  name_len += (mode_ext) ? strlen(fname_ext) : strlen(DEFAULT_EXT);

  fname_out = (char *)malloc(sizeof(char)*(name_len + 1));
  if (fname_out == NULL) {
    fprintf(stderr,"Couldn't allocate memory for fname_out\n");
    exit(1);
  }
  for (c = 0; c < name_len + 1; c++)
    fname_out[c] = '\0';

  strcpy(fname_out,fname_base);
  if (mode_suffix) {
    strcat(fname_out,"_"); strcat(fname_out,fname_suffix);
  }
  strcat(fname_out,"_");
  strcat(fname_out,comment_id);
  strcat(fname_out,"."); 
  strcat(fname_out,fname_ext); 

  return fname_out;
}

/* Simple version of the one defined in misc.c */
char *conv_out(char *buf_in)
{
#ifdef USE_ICONV
  int buf_len_mid0;
  size_t buf_len_mid, buf_len_out;
  char *b;
#ifdef ICONV_CONST
  const char *bm_tmp;
#else
  char *bm_tmp;
#endif
#endif
#ifdef USE_CRLF
  char *bi,*bm;
  int l;
#endif

#ifdef USE_CRLF
  /* Conversion of the line separaters: LF -> CR+LF */
  *buf_mid = '\0';
  bi = buf_in;
  bm = buf_mid;
  l = 0;
  while (*bi != '\0') {
    if (l >= MAX_OUT_LEN) {
      fprintf(stderr,"Too long line to output: %s",buf_in);
      exit(1);
    }
    if (*bi == '\n') {
      *bm = '\r';
      *(bm+1) = '\n';
      bm += 2;
      l += 2;
    } else {
      *bm = *bi;
      bm++;
      l++;
    }
    bi++;
  }
  *bm = '\0';
#else
  strncpy(buf_mid,buf_in,MAX_OUT_LEN);
#endif

#ifdef USE_ICONV
  if (!mode_conv_out) return buf_mid;

  *buf_out = '\0';
  b = buf_out;
  bm_tmp = buf_mid;
  buf_len_mid0 = strlen(buf_mid);
  buf_len_mid = (size_t)buf_len_mid0;
  buf_len_out = (size_t)MAX_OUT_LEN;
  if (iconv(cd_out,&bm_tmp,&buf_len_mid,&b,&buf_len_out) == (size_t)-1) {
    fprintf(stderr,"Couldn't convert: %s (errno=%d)\n",buf_mid,errno);
    exit(1);
  }
  buf_out[MAX_OUT_LEN - buf_len_out] = '\0';
#else
  return buf_mid;
#endif

  return buf_out;
}

void read_input(void)
{
  int line_c;
  FILE *fp_in = NULL;
  FILE *fp_out = NULL;
  char *fname_out;
  char *comment_id;
  char has_out;
  char printable;
  char *buf;
  char *buf_conv;
#ifdef USE_ICONV
  char *bb;
  size_t buf_len_in, buf_len_out;
#  ifdef ICONV_CONST
  const char *buf_tmp = NULL;
#  else
  char *buf_tmp = NULL;
#  endif
#endif
  char *s;

  buf = (char *)malloc(sizeof(char)*(MAX_LINE_LEN + 1));
  if (buf == NULL) {
    fprintf(stderr,"Couldn't allocate memory for buf\n");
    exit(1);
  }

#ifdef USE_ICONV
  buf_conv = (char *)malloc(sizeof(char)*(MAX_LINE_LEN + 1));
  if (buf_conv == NULL) {
    fprintf(stderr,"Couldn't allocate memory for buf\n");
    exit(1);
  }
#endif

  if (mode_stdin) {
    fp_in = stdin;
  }
  else {
    fp_in = fopen(fname_input,"rb");
    if (fp_in == NULL) {
      fprintf(stderr, "Couldn't open %s\n", fname_input);
      exit(1);
    }
  }

  line_c = 0;
  has_out = 0;
  printable = 0;
  while (fgets(buf,MAX_LINE_LEN,fp_in) != NULL) {
    line_c++;

#ifdef USE_ICONV
    if (mode_conv_in) {
      buf_tmp = buf;
      bb = buf_conv;
      buf_len_in  = strlen(buf)+1;
      buf_len_out = MAX_LINE_LEN+1;
      if (iconv(cd_in, &buf_tmp, &buf_len_in, &bb, &buf_len_out) == (size_t)-1) {
	fprintf(stderr,"[line %d] Couldn't convert to the internal charcode (errno=%d)\n",
		line_c,errno);
	exit(1);
      }
    } else {
      buf_conv = buf;
    }
#else
    buf_conv = buf;
#endif
    
    comment_id = get_comment_id(buf_conv);

    if (comment_id != NULL) {

      if (mode_list) {
	fprintf(stdout,"%s\n",comment_id);
	free(comment_id);
	continue;
      }

      if (!mode_stdout && has_out) fclose(fp_out);

      if (!mode_id ||
	  (mode_match == MATCH_INFIX  && strstr(comment_id,id_str) != NULL) ||
	  (mode_match == MATCH_EXACT  && strcmp(comment_id,id_str) == 0)    ||
	  (mode_match == MATCH_PREFIX &&
              (s = strstr(comment_id,id_str)) != NULL && comment_id == s) ||
	  (mode_match == MATCH_SUFFIX &&
              (s = strstr(comment_id,id_str)) != NULL && strlen(id_str) == strlen(s))) {

	if (mode_stdout) {
	  fputs(conv_out(buf_conv),stdout);
	  has_out = 0;
	}
	else {
	  fname_out = build_out_filename(comment_id);
	  
	  fprintf(stderr,"Output: %s\n",fname_out);
	  
	  fp_out = fopen(fname_out,"wb");
	  if (fp_out == NULL) {
	    fprintf(stderr,"Couldn't open %s\n", fname_out);
	    exit(1);
	  }
	  
	  has_out = 1;
	  fputs(conv_out(buf_conv),fp_out);
	  free(fname_out);
	}

	printable = 1;
      }
      else {
	has_out = 0;
	printable = 0;
      }

      free(comment_id);
    }
    else if (printable) {
      if (has_out)
	fputs(conv_out(buf_conv),fp_out);
      else
	fputs(conv_out(buf_conv),stdout);
    }

  }

  if (!mode_stdout && has_out) fclose(fp_out);
  if (!mode_stdin) fclose(fp_in);
}

int main(int argc, char *argv[])
{
  initialize_modes();
  initialize_buffer();
#ifdef USE_ICONV
  initialize_iconv();
#endif

  parse_args(argc, argv);

  read_input();

  release_buffer();
#ifdef USE_ICONV  
  iconv_close(cd_in);
  iconv_close(cd_out);
#endif

  return 0;
}
