#include "nbc_def.h"
#include "nbc_ext.h"

#if !defined(LANG) || defined(LANG_EN)

char usage[] =
    "\n<<<<  " TOOL_NAME ": A Toolkit for Naive Bayes Clustering (version " VERSION " / ML+MAP) >>>>\n\n"
    "[Usage] nbc -f basename -k #classes -x #attributes [option...]\n"
    "  Parameters:\n"
    "    -f Prefix of input/output files         [Required]\n"
    "    -k Number of classes or Kmin:Kmax:Kstep [Required unless -R or -H is given]\n"
    "    -x Number of attributes                 [Required]\n"
    "    -s Suffix of output files               [Default: No suffix]\n"
    "    -c Smoothing constant for class parameters\n"
    "         (auto:AutoClass, other positive float: as-is) [Default: No smoothing]\n"
    "    -w Smoothing constants for attribute parameters\n"
    "         (auto:AutoClass, other positive float: as-is) [Default: No smoothing]\n"
    "    -V Minimum smoothing constant for BDeu  [Default: BDeu is disabled]\n"
    "    -I Initialization method for class (random:Random, noisy_u:Noisy-uniform)\n"
    "                                            [Default: Noisy-uniform]\n"
    "    -J Initialization method for attributes\n"
    "         (random:Random, noisy_u:Noisy-uniform: v_freq:Value-frequency)\n"
    "                                            [Default: Noisy-uniform]\n"
    "    -e Threshold in judging convergence     [Default: 1.0e-6]\n"
    "    -m Max number of EM iterations          [Default: Iterate EM until convergence]\n"
    "    -t Clustering type (mp: membership prob, wodd: weighted log odds ratio)\n"
    "    -g Information criterion (bic:BIC, autoclass:AUTOCLASS)      [Default: BIC]\n"
    "         [NOTE] AUTOCLASS is available only when smoothing is enable (with -c and -w)\n"
    "    -r Random seed                          [Default: a return value of time()]\n"
    "    -n Number of initial parameter settings in random restart    [Default: No restart]\n"
    "    -l Max number of preliminary EM iterations in random restart\n"
    "                                            [Default: Iterate EM until convergence]\n"
    "    -z Number of attribute values highly relevant to a class (Ranking is enabled)\n"
    "                                            [Default: Ranking is disabled] \n"
    "    -Q Measure of relevance between a class and an attribute value\n"
    "         (mp:P(k|x), wodd:weighted log odd ratio)           [Default: mp] \n"
    "    -o Max number of split-merge operations (Enable SMEM)   [Default: SMEM is disabled]\n"
    "    -a Max number of SM candidates to be examined           [Default: No limit]\n"
    "    -G Max number of split candidate to be examined for each pair being merged\n"
    "                                                            [Default: No limit]\n"
    "    -U Lower limit of prob. of the class allowed to be split/merged [Default: No limit]\n"
    "    -C Lower limit of prob. of the splited/merged class     [Default: No limit]\n"
    "    -N Lower limit of improvement in spilt-merge oparations\n"
    "                               [Default: Threshold of convergence (given by -e option)]\n"
#ifdef USE_ICONV
#ifdef CHARCODE
    "    -p Input character charcode (run 'iconv -l' for available charcodes)  [Default: " CHARCODE "]\n"
    "    -q Output character charcode (run 'iconv -l' for available charcodes) [Default: " CHARCODE "]\n"
#else
    "    -p Input character charcode (run 'iconv -l' for available charcodes)  [Default: " CHARCODE_INTERNAL "]\n"
    "    -q Output character charcode (run 'iconv -l' for available charcodes) [Default: " CHARCODE_INTERNAL "]\n"
#endif
#endif
    "  Modes:\n"
    "    -R Disable EM learning and read paramters from a file\n"
    "    -H Read hyper-parameters from a file\n"
    "    -E Evalute the obtained clusters\n"
    "    -S Output the messages to the files         [Default: to stderr]\n"
    "    -D Compute and output the membership distributions\n"
    "    -d Compute and output the dissimilarities\n"
    "    -i Assume the input data is in CSV form     [Default: TAB-separated form]\n"
    "    -j Output the computed results in CSV form  [Default: TAB-separated form]\n"
    "    -v Output the temporarily computed results to log file\n"
    "    -F Normalize Jmerge\n"
    "    -M Save memory for normal EM (though it can be slower)\n"
    "    -Y Print attribute values in the appearance order   [Default: lexicographical order]\n"
    "    -h Print this message\n\n";

void print_usage(FILE *fp)
{
  fprintf(fp,conv_msg(usage));
}

void print_log_likelihood(FILE *fp, int k, int num_iterate)
{
  int m;

  if (mode_smooth) {
    conv_fprintf(fp,"#### ID:LOGLIKE-K%d\n#### Number of classes~, Number of updates~, Log likelihood~, Log prior prob.~, Log a posteriori prob.\n####\n",k);
    for (m=0; m<num_iterate; m++) {
      conv_fprintf(fp,"%d,%d,%.12Lf,%.12Lf,%.12Lf\n",
	      num_class,m,log_likelihood[m],log_prior[m],lambda[m]);
    }
  } else {
    conv_fprintf(fp,"#### ID:LOGLIKE-K%d\n#### Number of classes~, Number of updates~, Log likelihood\n####\n",k);
    for (m=0; m<num_iterate; m++) {
      conv_fprintf(fp,"%d,%d,%.12Lf\n",num_class,m,log_likelihood[m]);
    }
  }

  fflush(fp);
}

void print_ic(FILE *fp)
{
  int k,u;

  conv_fprintf(fp,"#### ID:SCORE\n#### Number of classes~, BIC~, AUTOCLASS\n"
	       "#### Selected criterion = %s~, optimal number of classes = %d\n####\n",
	  ic_type_str,num_class_opt);
  u = 0;
  for (k = kmin; k <= kmax; k += kstep) {
    if (mode_smooth) {
      conv_fprintf(fp,"%d,%.12Lf,%.12Lf\n",k,bic[u],autoclass[u]);
    } else {
      conv_fprintf(fp,"%d,%.12Lf,N/A\n",k,bic[u]);
    }
    u++;
  }

  fflush(fp);
}

void print_seed(FILE *fp)
{
  int k,u,r;

  if (mode_search_num_class) {
    u = 0;
    for (k = kmin; k <= kmax; k += kstep) {
      if (mode_smooth) {
	conv_fprintf(fp,"#### ID:SEED-K%d\n#### Number of classes~, Restart(best==%d)~, Random seed~, Number of updates~, Log a posteriori prob.\n####\n",k,log_r_best[u]);
      } else {
        conv_fprintf(fp,"#### ID:SEED-K%d\n#### Number of classes~, Restart(best=%d)~, Random seed~, Number of updates~, Log likelihood\n####\n",k,log_r_best[u]);
      }
      
      for (r = 0; r < num_init; r++) {
	if (mode_set_seed) {
	  conv_fprintf(fp,"%d,%d,N/A,%d,%.12Lf\n",
		  k,r,
		  log_num_iterate[u][r],log_final_lambda[u][r]);
	} else {
	  conv_fprintf(fp,"%d,%d,%u,%d,%.12Lf\n",
		  k,r,
		  log_seed[u][r],log_num_iterate[u][r],log_final_lambda[u][r]);
	}
      }
      u++;
    }

  } else {
    
    if (mode_smooth) {
      conv_fprintf(fp,"#### ID:SEED\n#### Number of classes~, Restart(best=%d)~, Random seed~, Number of updates~, Log a posteriori prob.\n####\n",log_r_best[0]);
    } else {
      conv_fprintf(fp,"#### ID:SEED\n#### Number of classes~, Restart(best=%d)~, Random seed~, Number of updates~, Log likelihood\n####\n",log_r_best[0]);
    }
    for (r = 0; r < num_init; r++) {
      if (mode_set_seed) {
	conv_fprintf(fp,"%d,%d,N/A,%d,%.12Lf\n",
		num_class,r,
		log_num_iterate[0][r],log_final_lambda[0][r]);
      } else {
	conv_fprintf(fp,"%d,%d,%u,%d,%.12Lf\n",
		num_class,r,
		log_seed[0][r],log_num_iterate[0][r],log_final_lambda[0][r]);
      }
    }
  }

  fflush(fp);
}

void print_decrease(FILE *fp, int r, int m,
		    FLOAT_TYPE pre_lambda, FLOAT_TYPE post_lambda)
{
  if (mode_smooth) {
    conv_fprintf(fp,"#### ID:ERR_MSG-DECREASE\n#### Log posterior prob. decreased!\n####\n");
  } else {
    conv_fprintf(fp,"#### ID:ERR_MSG-DECREASE\n#### Log-likelihood decreased!\n####\n");
  }
  conv_fprintf(fp,"NumberOfClasses,%d\n",num_class);
  conv_fprintf(fp,"Restart,%d\n",r);
  conv_fprintf(fp,"NumberOfUpdates,%d\n",m);
  if (mode_smooth) {
    conv_fprintf(fp,"LogPostProbBeforeUpdate,%.15Lf\n",pre_lambda);
    conv_fprintf(fp,"LogPostProbAfterUpdate,%.15Lf\n",post_lambda);
  } else {
    conv_fprintf(fp,"LogLikelihoodBeforeUpdate,%.15Lf\n",pre_lambda);
    conv_fprintf(fp,"LogLikelihoodAfterUpdate,%.15Lf\n",post_lambda);
  }
}

void print_time(FILE *fp)
{
  conv_fprintf(fp,"#### ID:CPUTIME\n#### CPU time (sec)\n####\n");
  conv_fprintf(fp,"Read,%.6Lf\n",(etime_read - stime_read));
  if (mode_em)
    conv_fprintf(fp,"EM,%.6Lf\n",(etime_em - stime_em));
  if (mode_memp)
    conv_fprintf(fp,"MemberDistrib,%.6Lf\n",(etime_memp - stime_memp));
  if (mode_dist) {
    conv_fprintf(fp,"KL-Divergence,%.6Lf\n",(etime_kl_dist - stime_kl_dist));
  }
  fflush(fp);

  conv_fprintf(fp,"#### ID:LAPTIME\n#### Elapsed time (sec)\n####\n");
  conv_fprintf(fp,"Read,%.6Lf\n",(etime_lap_read - stime_lap_read));
  if (mode_em)
    conv_fprintf(fp,"EM,%.6Lf\n",(etime_lap_em - stime_lap_em));
  if (mode_memp)
    conv_fprintf(fp,"MemberDistrib,%.6Lf\n",(etime_lap_memp - stime_lap_memp));
  if (mode_dist) {
    conv_fprintf(fp,"KL-Divergence,%.6Lf\n",(etime_lap_kl_dist - stime_lap_kl_dist));
  }
  fflush(fp);
}

void print_jmerge(FILE *fp)
{
  int k1,k2;
  conv_fprintf(fp,"#### ID:JMERGE\n#### Class 1~, Class 2~, Jmerge\n####\n");
  for (k1=0; k1<num_class; k1++) {
    for (k2=0; k2<num_class; k2++) {
      conv_fprintf(fp,"%d,%d,%.9Lf\n",k1,k2,jmerge[k1][k2]);
    }
  }
}

void print_jsplit(FILE *fp)
{
  int k;
  conv_fprintf(fp,"#### ID:JSPLIT\n#### Class 1~, Class 2~, Jsplit\n####\n");
  for (k=0; k<num_class; k++) {
    conv_fprintf(fp,"%d,%.9Lf\n",k,jsplit[k]);
  }
}

void print_log_likelihood_sm(FILE *fp, int k)
{
  int t,m;

  if (mode_smooth) {
    conv_fprintf(fp,"#### ID:LOGLIKE-SMEM-K%d\n#### Number of classes~, Number of split-merge operations~, Number of updates~, Log likelihood~, Log prior prob.~, Log a posteriori prob.\n####\n",k);
    for (t=0; t<(num_sm+1); t++) {
      for (m=0; m<num_iterate_sm_full[t]; m++) {
	conv_fprintf(fp,"%d,%d,%d,%.12Lf,%.12Lf,%.12Lf\n",
		k,t,m,
		log_likelihood_sm[t][m],
		log_prior_sm[t][m],
		lambda_sm[t][m]);
      }
    }
  } else {
    conv_fprintf(fp,"#### ID:LOGLIKE-SMEM-K%d\n#### Number of classes~, Number of split-merge operations~, Number of updates~, Log likelihood\n####\n",k);
    for (t=0; t<(num_sm+1); t++) {
      for (m=0; m<num_iterate_sm_full[t]; m++) {
	conv_fprintf(fp,"%d,%d,%d,%.12Lf\n",
		k,t,m,log_likelihood_sm[t][m]);
      }
    }
  }

}

void print_num_result_sm(FILE *fp, int k)
{
  int t;

  conv_fprintf(fp,"#### ID:SMEM-OP-STAT-K%d\n#### Number of classes~, Number of split-merge operations~, Number of rejected candidates~, Number of ignored candidates\n####\n",k);
  for (t=0; t<num_sm; t++) {
    conv_fprintf(fp,"%d,%d,%d,%d\n",k,t,num_result_rejected[t],num_result_skipped[t]);
  }
}

void print_num_iterate_sm(FILE *fp, int k)
{
  int t;

  conv_fprintf(fp,"#### ID:SMEM-EM-STAT-K%d\n#### Number of classes~, Number of split-merge operations~, Number of updates in partial EM~, Number of updates in full EM\n####\n",k);
  for (t=0; t<(num_sm+1); t++) {
    if (t==0) {
      conv_fprintf(fp,"%d,%d,N/A,%d\n",k,t,num_iterate_sm_full[t]);
    } else {
      conv_fprintf(fp,"%d,%d,%d,%d\n",
	      k,t,num_iterate_sm_part[t],
	      (num_iterate_sm_full[t]-num_iterate_sm_part[t]));
    }
  }

}

void print_settings(FILE *fp)
{
  int v;

  conv_fprintf(fp,"#### ID:SETTING\n#### Settings:\n####\n");
  conv_fprintf(fp,"Version,%s\n",VERSION);
  conv_fprintf(fp,"CaseFile(input),%s\n",fname_freq);
  conv_fprintf(fp,"ParameterFile(output),%s\n",fname_param);
  conv_fprintf(fp,"MemberDistribFile(output),%s\n",fname_memp);
  conv_fprintf(fp,"DistanceFile(output),%s\n",fname_dist);
  conv_fprintf(fp,"LogFile(output),%s\n",fname_log);
  if (mode_cluster) {
    conv_fprintf(fp,"ClusterFile(output),%s\n",fname_cluster);
  }
  if (mode_rank) {
    conv_fprintf(fp,"RankFile(output),%s\n",fname_rank);
  }
  if (mode_rank_sm) {
    conv_fprintf(fp,"SmemRankLogFile(output),%s\n",fname_rank_sm);
  }
  if (!mode_search_num_class) {
    conv_fprintf(fp,"NumberOfClasses,%d\n",num_class);
  } else {
    conv_fprintf(fp,"NumberOfClasses(Min),%d\n",kmin);
    conv_fprintf(fp,"NumberOfClasses(Max),%d\n",kmax);
    conv_fprintf(fp,"NumberOfClasses(Step),%d\n",kstep);
  }
  conv_fprintf(fp,"NumberOfAttributes,%d\n",tuple_size);
  conv_fprintf(fp,"SmoothingMode,%d\n",mode_smooth);
  conv_fprintf(fp,"SmoothingMode(Strict),%d\n",mode_smooth_strict);
  conv_fprintf(fp,"MemberDistribMode,%d\n",mode_memp);
  conv_fprintf(fp,"DistanceMode,%d\n",mode_dist);
  conv_fprintf(fp,"OptimalClassSearchMode,%d\n",mode_search_num_class);
  conv_fprintf(fp,"CsvInputMode,%d\n",mode_csv_in);
  conv_fprintf(fp,"CsvOutputMode,%d\n",mode_csv_out);
#ifdef USE_ICONV  
  conv_fprintf(fp,"InputCharMode,%s\n",charcode_in);
  conv_fprintf(fp,"OutputCharMode,%s\n",charcode_out);
#endif
  conv_fprintf(fp,"MemorySavingMode,%d\n",mode_nomem);
  conv_fprintf(fp,"RankingMode,%d\n",mode_rank);
  conv_fprintf(fp,"RankingMode(SMEM),%d\n",mode_rank_sm);
  conv_fprintf(fp,"NumberOfRankedAttrituteValues,%d\n",num_rank);
  conv_fprintf(fp,"NumberOfRankedAttrituteValues(SMEM),%d\n",num_rank_sm);
  conv_fprintf(fp,"RankingScore,%s\n",rel_m_str);
  if (mode_auto_smooth_c == MODE_SMOOTH_AUTO) {
    conv_fprintf(fp,"ClassSmoothingConstant,AutoClass\n");
  } else if (mode_auto_smooth_c == MODE_SMOOTH_AUTO) {
    conv_fprintf(fp,"ClassSmoothingConstant,BDeu\n");
  } else {
    conv_fprintf(fp,"ClassSmoothingConstant,%.6Lf\n",smooth_c);
  }
  for (v=0; v<tuple_size; v++) {
    if (mode_auto_smooth_w[v] == MODE_SMOOTH_AUTO) {
      conv_fprintf(fp,"AttributeSmoothingConstant%d,AutoClass\n",v);
    } else if (mode_bdeu) {
      conv_fprintf(fp,"AttributeSmoothingConstant%d,BDeu\n",v);
    } else {
      conv_fprintf(fp,"AttributeSmoothingConstant%d,%.6Lf\n",v,smooth_w[v]);
    }
  }
  conv_fprintf(fp,"MaximumNumberOfEmIterations,%d\n",max_iterate);
  conv_fprintf(fp,"ConvergenceThreshold,%.3Le\n",epsilon);
  conv_fprintf(fp,"InitializationMethod(Class),%s\n",init_type_str_class);
  conv_fprintf(fp,"InitializationMethod(Attribute),%s\n",init_type_str_word);
  if (mode_search_num_class) {
    conv_fprintf(fp,"InformationCriterion,%s\n",ic_type_str);
  }
  if (num_init > 1) {
    conv_fprintf(fp,"NumberOfRestarts,%d\n",num_init);
    conv_fprintf(fp,"MaximumNumberOfEmIterationsWithRestart,%d\n",max_init_iterate);
  }
  if (num_sm > 0) {
    conv_fprintf(fp,"NumberOfSplitMergeOperations,%d\n",num_sm);
    conv_fprintf(fp,"MaximumNumberOfSplitMergeCandidates,%d\n",max_candidate);
    conv_fprintf(fp,"JmergeNormalizationMode,%d\n",mode_norm);
    conv_fprintf(fp,"MinimumClassProbabilityBeforeSplitMerge(0:NoBound),%.6Lf\n",
	    min_param_c);
    conv_fprintf(fp,"MinimumClassProbabilityAfterSplitMerge(0:NoBound),%.6Lf\n",
	    min_new_param_c);
    conv_fprintf(fp,"MinimumImprovementBySplitMerge,%.6Lf\n",
	    min_smem_improve);
  }
  conv_fprintf(fp,"RandomSeed,%d\n",rand_seed);
#ifndef NO_LDOUBLE
  conv_fprintf(fp,"ByteSize(double),%d\n",sizeof(double));
  conv_fprintf(fp,"ByteSize(long-double),%d\n",sizeof(FLOAT_TYPE));
#endif

  fflush(fp);
}

void print_att_list(FILE *fp)
{
  int v,i,j;

  for (v=0; v<tuple_size; v++) {
    conv_fprintf(fp,"#### ID:ATTVAL\n#### ID for the attribute value~, Attr. value x%d\n####\n",v,v);
    for (i=0; i<num_word[v]; i++) {
      j = print_order[v][i];
      conv_fprintf(fp,"%d,%s\n",j,word_list[v][j]);
    }
  }

  fflush(fp);
}

void print_att_freq_list(FILE *fp)
{
  int v,i,j;

  for (v=0; v<tuple_size; v++) {
    conv_fprintf(fp,"#### ID:ATTVAL-FREQ\n#### ID for the attribute value~, Attr. value x%d~, Number of occurrences N(x%d)\n####\n",v,v);
    for (i=0; i<num_word[v]; i++) {
      j = print_order[v][i];
      conv_fprintf(fp,"%d,%s,%d\n",j,word_list[v][j],word_freq_list[v][j]);
    }
  }

  fflush(fp);
}

void print_case_freq_list(FILE *fp)
{
  int v,e;

  conv_fprintf(fp,"#### ID:CASE-FREQ\n#### ");
  if (mode_eval) conv_fprintf(fp,"Answer class~, ");
  if (tuple_size > 2) {
    conv_fprintf(fp,"Attr. value x1~, x2~, ...~, x%d~, ",tuple_size);
    conv_fprintf(fp,"Number of occurrences N(x1~,x2~,...~,x%d)\n####\n",tuple_size);
  } else {
    conv_fprintf(fp,"Attr. value x1~, x2~, ");
    conv_fprintf(fp,"Number of occurrences N(x1~,x2)\n####\n");
  }

  for (e=0; e<num_entry; e++) {
    if (mode_eval)
      conv_fprintf(fp,"%s,",ans_class_list[freq_list[e].ans_id]);
    for (v=0; v<tuple_size; v++)
      conv_fprintf(fp,"%s,",word_list[v][freq_list[e].id[v]]);
    conv_fprintf(fp,"%d\n",
	    freq_list[e].freq);
  }

  fflush(fp);
}

void print_ans_class_list(FILE *fp)
{
  int k;
  conv_fprintf(fp,"#### ID:ANS-CLASS\n#### Answer class (for evaluation)~, "
	       "Answer class name\n####\n");
  for (k=0; k<num_ans_class; k++) {
    conv_fprintf(fp,"%d,%s\n",k,ans_class_list[k]);
  }
}

void print_smooth(FILE *fp, int k)
{
  int v;

  conv_fprintf(fp,"#### ID:AUTO-SMOOTH-CLASS-K%d\n"
	       "#### Number of classes~, Class smoothing constant (based on autoclass/BDeu)\n####\n",k);
  conv_fprintf(fp,"%d,%.12Lf\n",k,smooth_c);

  conv_fprintf(fp,"#### ID:AUTO-SMOOTH-ATT-K%d\n"
	       "#### Number of classes~, Attr. ID~, Attr. smoothing constant (based on autoclass/BDeu)\n####\n",k);
  for (v=0; v<tuple_size; v++) {
    conv_fprintf(fp,"%d,%d,%.12Lf\n",k,v,smooth_w[v]);
  }
}

void print_size(FILE *fp)
{
  int k,v;
  int total_num_word;

  conv_fprintf(fp,"#### ID:VALUESIZE\n#### Attr. ID~, Number of attribute values\n####\n");
  for (v=0; v<tuple_size; v++)
    conv_fprintf(fp,"%d,%d\n",v,num_word[v]);

  conv_fprintf(fp,"#### ID:DATASIZE\n#### Total number of cases~, Number of entries (original)~, Number of entries (multiple occurences compressed)\n####\n");
  conv_fprintf(fp,"%d,%d,%d\n",total_freq,num_entry_pre,num_entry);

  conv_fprintf(fp,"#### ID:PARAMSIZE\n#### Number of classes~, Number of class parameters~, ");
  if (tuple_size > 2) {
    conv_fprintf(fp,"Number of parameters for attribute 1~, Number of parameters for attribute 2~, ...~, Number of parameters for attribute %d~, ",(tuple_size+1));
  } else {
    conv_fprintf(fp,"Number of parameters for attribute 1~, Number of parameters for attribute 2~, ",(tuple_size+1));
  }
  conv_fprintf(fp,"Total number of parameters\n####\n");
  if (mode_search_num_class) {
    for (k = kmin; k <= kmax; k += kstep) {
      conv_fprintf(fp,"%d,%d",k,(k-1));
      for (v=0; v<tuple_size; v++)
	conv_fprintf(fp,"%d,",k*(num_word[v] - 1));
      total_num_word = 0;
      for (v=0; v<tuple_size; v++)
	total_num_word += num_word[v];
      conv_fprintf(fp,"%d\n",(k*(total_num_word - 1) - 1));
    }
  } else {
    k = num_class;
    conv_fprintf(fp,"%d,%d,",k,(k-1));
    for (v=0; v<tuple_size; v++)
      conv_fprintf(fp,"%d,",k*(num_word[v] - 1));
    total_num_word = 0;
    for (v=0; v<tuple_size; v++)
      total_num_word += num_word[v];
    conv_fprintf(fp,"%d\n",(k*(total_num_word - 1) - 1));
  }

  fflush(fp);
}

void print_param(FILE *fp)
{
  int k,v,i,j;

  conv_fprintf(fp,"#### ID:PARAM-CLASS\n#### Class k~, Class parameter P(k)\n####\n");
  for (k=0; k<num_class; k++) {
    conv_fprintf(fp,"%d,%.12Lf\n",k,param_c[k]);
  }

  for (v=0; v<tuple_size; v++) {
    conv_fprintf(fp,"#### ID:PARAM-ATT%d\n#### Class k~, Attr. value x%d~, Attr. parameter P(x%d|k)\n####\n",v,v,v);

    for (k=0; k<num_class; k++) {
      for (i=0; i<num_word[v]; i++) {
        j = print_order[v][i];
	conv_fprintf(fp,"%d,%s,%.12Lf\n",k,word_list[v][j],param_w[k][v][j]);
      }
    }
  }

  fflush(fp);
}

void print_complete_prob(FILE *fp)
{
  int k,v,e;

  conv_fprintf(fp,"#### ID:COMP\n#### Class k~, ");
  for (v=0; v<tuple_size; v++)
    conv_fprintf(fp,"Attr. value x%d~, ",v);
  if (tuple_size > 2) {
    conv_fprintf(fp,"Complete data prob. P(k,x1~,x2~,...~,x%d)\n####\n",tuple_size);
  } else {
    conv_fprintf(fp,"Complete data prob. P(k,x1~,x2)\n####\n");
  }

  for (k=0; k<num_class; k++) {
    for (e=0; e<num_entry; e++) {
      conv_fprintf(fp,"%d,",k);
      for (v=0; v<tuple_size; v++)
	conv_fprintf(fp,"%s,",word_list[v][freq_list[e].id[v]]);
      conv_fprintf(fp,"%.12Lf\n",prob_comp[k][e]);
    }
  }
  fflush(fp);
}

void print_incomplete_prob(FILE *fp)
{
  int v,e;

  conv_fprintf(fp,"#### ID:INCOMP\n#### ");
  for (v=0; v<tuple_size; v++) {
    conv_fprintf(fp,"Attr. value x%d~, ",v);
  }
  if (tuple_size > 2) {
    conv_fprintf(fp,"Incomplete data prob. P(x1~,x2~,...~,x%d)\n####\n",tuple_size);
  } else {
    conv_fprintf(fp,"Incomplete data prob. P(x1~,x2)\n####\n");
  }

  for (e=0; e<num_entry; e++) {
    for (v=0; v<tuple_size; v++)
      conv_fprintf(fp,"%s,",word_list[v][freq_list[e].id[v]]);
    conv_fprintf(fp,"%.12Lf\n",prob_incomp[e]);
  }

  fflush(fp);
}

void print_conditional_prob(FILE *fp)
{
  int k,e,v;

  conv_fprintf(fp,"#### ID:COND\n#### ");
  for (v=0; v<tuple_size; v++) {
    conv_fprintf(fp,"Attr. value x%d~, ",v);
  }
  if (tuple_size > 2) {
    conv_fprintf(fp,"Class k~, Conditional prob. P(k|x1~,x2~,...~,x%d)\n####\n",tuple_size);
  } else {
    conv_fprintf(fp,"Class k~, Conditional prob. P(k|x1~,x2)\n####\n");
  }

  for (e=0; e<num_entry; e++) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size; v++)
	conv_fprintf(fp,"%s,",word_list[v][freq_list[e].id[v]]);
      conv_fprintf(fp,"%d,%.12Lf\n",k,prob_cond[k][e]);
    }
  }

  fflush(fp);
}

void print_expectation(FILE *fp)
{
  int k,v,i,j;

  conv_fprintf(fp,"#### ID:EXPECT-CLASS\n#### Class k~, Expected occurrences E[k]\n####\n");
  for (k=0; k<num_class; k++) {
    conv_fprintf(fp,"%d,%.12Lf\n",k,expect_c[k]);
  }

  for (v=0; v<tuple_size; v++) {

    conv_fprintf(fp,"#### ID:EXPECT-ATT%d\n#### Class k~, Attr. value x%d~, Expected occurrences E[x%d|k]\n####\n",v,v,v);
    for (k=0; k<num_class; k++) {
      for (i=0; i<num_word[v]; i++) {
        j = print_order[v][i];
	conv_fprintf(fp,"%d,%s,%.12Lf\n",k,word_list[v][j],expect_w[k][v][j]);
      }
    }
  }

  fflush(fp);
}

void print_member_prob(FILE *fp)
{
  int k,v,i,j,e;

  conv_fprintf(fp,"#### ID:MEMBER-CASE\n#### Case ID~, ");
  if (mode_eval) conv_fprintf(fp,"Answer class~, ");
  if (tuple_size > 2) {
    conv_fprintf(fp,"Attr. x1~, ...~, Attr. x%d~, ",tuple_size);
  }
  else {
    conv_fprintf(fp,"Attr. x1~, Attr. x2~, ");
  }
  conv_fprintf(fp,"Class k~, Membership prob. P(k|x)\n####\n");
  for (e=0; e<num_entry; e++) {
    for (k=0; k<num_class; k++) {
      conv_fprintf(fp,"%d,",e);
      if (mode_eval) conv_fprintf(fp,"%s,",ans_class_list[freq_list[e].ans_id]);
      for (v=0; v<tuple_size; v++)
	conv_fprintf(fp,"%s,",word_list[v][freq_list[e].id[v]]);
      conv_fprintf(fp,"%d,",k);
      conv_fprintf(fp,"%.12Lf\n",prob_case_mem[k][e]);
    }
  }

  for (v=0; v<tuple_size; v++) {

    conv_fprintf(fp,"#### ID:MEMBER-ATT%d\n#### Attr. value x%d~, Class k~, Membership prob. P(k|x%d)\n####\n",v,v,v);

    for (i=0; i<num_word[v]; i++) {
      j = print_order[v][i];
      for (k=0; k<num_class; k++)
	conv_fprintf(fp,"%s,%d,%.12Lf\n",word_list[v][j],k,prob_att_mem[k][v][j]);
    }
  }

  fflush(fp);
}

void print_cluster(FILE *fp)
{
  int k,t,i,v,e;

  switch (cl_type) {

  case CL_MP:
    conv_fprintf(fp,"#### ID:CLUSTER-CASE-BY-CASE\n#### Case ID~, ");
    if (mode_eval) conv_fprintf(fp,"Answer class~, ");
    conv_fprintf(fp,"Cluster k~, ");
    if (tuple_size > 2) {
      conv_fprintf(fp,"Attr. x1~, ...~, Attr. x%d~, ",tuple_size);
    }
    else {
      conv_fprintf(fp,"Attr. x1~, Attr. x2~, ");
    }
    conv_fprintf(fp,"Membership prob. P(k|x)\n####\n");
    for (e=0; e<num_entry; e++) {
      k = case_cluster_pred[e];
      conv_fprintf(fp,"%d,",e);
      if (mode_eval) conv_fprintf(fp,"%s,",ans_class_list[freq_list[e].ans_id]);
      conv_fprintf(fp,"%d,",k);
      for (v=0; v<tuple_size; v++)
	conv_fprintf(fp,"%s,",word_list[v][freq_list[e].id[v]]);
      conv_fprintf(fp,"%.12Lf\n",prob_case_mem[k][e]);
    }
    conv_fprintf(fp,"#### ID:CLUSTER-CASE-BY-CLUSTER\n#### Cluster k~, Case ID~, ");
    if (mode_eval) conv_fprintf(fp,"Answer class~, ");
    if (tuple_size > 2) {
      conv_fprintf(fp,"Attr. x1~, ...~, Attr. x%d~, ",tuple_size);
    }
    else {
      conv_fprintf(fp,"Attr. x1~, Attr. x2~, ");
    }
    conv_fprintf(fp,"Membership prob. P(k|x)\n####\n");
    for (k=0; k<num_class; k++) {
      for (t=0; t<case_cluster_freq[k]; t++) {
        e = case_cluster[k][t];
	conv_fprintf(fp,"%d,%d,",k,e);
	if (mode_eval) conv_fprintf(fp,"%s,",ans_class_list[freq_list[e].ans_id]);
        for (v=0; v<tuple_size; v++)
          conv_fprintf(fp,"%s,",word_list[v][freq_list[e].id[v]]);
        conv_fprintf(fp,"%.12Lf\n",prob_case_mem[k][e]);
      }
    }
    break;

  case CL_WODD:
    conv_fprintf(fp,"#### ID:CLUSTER-CASE-BY-CASE\n#### Case ID~, ");
    if (mode_eval) conv_fprintf(fp,"Answer class~, ");
    conv_fprintf(fp,"Cluster k~, ");
    if (tuple_size > 2) {
      conv_fprintf(fp,"Attr. x1~, ...~, Attr. x%d~, ",tuple_size);
    }
    else {
      conv_fprintf(fp,"Attr. x1~, Attr. x2~, ");
    }
    conv_fprintf(fp,"Membership prob. P(k|x)~, WLLR P(k|x)log{P(k|x)/P(k|not x)}\n####\n");
    for (e=0; e<num_entry; e++) {
      k = case_cluster_pred[e];
      conv_fprintf(fp,"%d,",e);
      if (mode_eval) conv_fprintf(fp,"%s,",ans_class_list[freq_list[e].ans_id]);
      conv_fprintf(fp,"%d,",k);
      for (v=0; v<tuple_size; v++)
	conv_fprintf(fp,"%s,",word_list[v][freq_list[e].id[v]]);
      conv_fprintf(fp,"%.12Lf,%.12Lf\n",
	      prob_case_mem[k][e],wodd_case[k][e]);
    }
    conv_fprintf(fp,"#### ID:CLUSTER-CASE-BY-CLUSTER\n#### Cluster k~, Case ID~, ");
    if (mode_eval) conv_fprintf(fp,"Answer class~, ");
    if (tuple_size > 2) {
      conv_fprintf(fp,"Attr. x1~, ...~, Attr. x%d~, ",tuple_size);
    }
    else {
      conv_fprintf(fp,"Attr. x1~, Attr. x2~, ");
    }
    conv_fprintf(fp,"Membership prob. P(k|x)~, WLLR P(k|x)log{P(k|x)/P(k|not x)}\n####\n");
    for (k=0; k<num_class; k++) {
      for (t=0; t<case_cluster_freq[k]; t++) {
        e = case_cluster[k][t];
	conv_fprintf(fp,"%d,%d,",k,e);
	if (mode_eval) conv_fprintf(fp,"%s,",ans_class_list[freq_list[e].ans_id]);
        for (v=0; v<tuple_size; v++)
          conv_fprintf(fp,"%s,",word_list[v][freq_list[e].id[v]]);
        conv_fprintf(fp,"%.12Lf,%.12Lf\n",
                prob_case_mem[k][e],wodd_case[k][e]);
      }
    }
    break;

  default:
    die("Internal error: Unsupported clustering method\n");
  }

  for (v=0; v<tuple_size; v++) {
    switch (cl_type) {

    case CL_MP:
      conv_fprintf(fp,"#### ID:CLUSTER-ATT%d-BY-ATT\n#### Cluster k~, Attr. x%d~, Membership prob. P(k|x%d)\n####\n",v,v,v,v);
      for (i=0; i<num_word[v]; i++) {
	k = att_cluster_pred[v][i];
	conv_fprintf(fp,"%d,%s,%.12Lf\n",
		k,word_list[v][i],prob_att_mem[k][v][i]);
      }
      conv_fprintf(fp,"#### ID:CLUSTER-ATT%d-BY-CLUSTER\n#### Cluster k~, Attr. x%d~, Membership prob. P(k|x%d)\n####\n",v,v,v,v);
      for (k=0; k<num_class; k++) {
	for (t=0; t<att_cluster_freq[v][k]; t++) {
	  i = att_cluster[v][k][t];
	  conv_fprintf(fp,"%d,%s,%.12Lf\n",
		  k,word_list[v][i],prob_att_mem[k][v][i]);
	}
      }
      break;

    case CL_WODD:
      conv_fprintf(fp,"#### ID:CLUSTER-ATT%d-BY-ATT\n#### Cluster k~, Attr. value x%d~, Membership prob. P(k|x%d)~, WLLR P(k|x%d)log{P(k|x%d)/P(k|not x%d)}\n####\n",v,v,v,v,v,v);
      for (i=0; i<num_word[v]; i++) {
	k = att_cluster_pred[v][i];
	conv_fprintf(fp,"%d,%s,%.12Lf,%.12Lf\n",
		k,word_list[v][i],prob_att_mem[k][v][i],wodd_att[k][v][i]);
      }
      conv_fprintf(fp,"#### ID:CLUSTER-ATT%d-BY-CLUSTER\n#### Cluster k~, Attr. value x%d~, Membership prob. P(k|x%d)~, WLLR P(k|x%d)log{P(k|x%d)/P(k|not x%d)}\n####\n",v,v,v,v,v,v);
      for (k=0; k<num_class; k++) {
	for (t=0; t<att_cluster_freq[v][k]; t++) {
	  i = att_cluster[v][k][t];
	  conv_fprintf(fp,"%d,%s,%.12Lf,%.12Lf\n",
		  k,word_list[v][i],prob_att_mem[k][v][i],wodd_att[k][v][i]);
	}
      }
      break;

    default:
      die("Internal error: Unsupported clustering method\n");
    }
  }
}

void print_cluster_freq(FILE *fp)
{
  int k,v;

  conv_fprintf(fp,"#### ID:CLUSTER-FREQ-CASE\n#### Cluster for cases~, Cluster size\n####\n");
  for (k=0; k<num_class; k++) {
    conv_fprintf(fp,"%d,%d\n",k,case_cluster_freq[k]);
  }

  for (v=0; v<tuple_size; v++) {
    conv_fprintf(fp,"#### ID:CLUSTER-FREQ-ATT%d\n#### Cluster for attribute %d~, Cluster size\n####\n",v,v);
    for (k=0; k<num_class; k++) {
      conv_fprintf(fp,"%d,%d\n",k,att_cluster_freq[v][k]);
    }
  }

  fflush(fp);
}

void print_rank_ml(FILE *fp)
{
  int k,k0,i,j,v,e;
  int num_show_case_rank;
  int *num_show_att_rank;

  num_show_att_rank = (int *)malloc(sizeof(int)*tuple_size);
  if (num_show_att_rank == NULL)
    die("Couldn't allocate memory for num_show_att_rank\n");

  num_show_case_rank = (num_rank > 0 && num_rank < num_entry) ? num_rank : num_entry;
  for (v=0; v<tuple_size; v++)
    num_show_att_rank[v] = (num_rank > 0 && num_rank < num_word[v]) ? num_rank : num_word[v];

  conv_fprintf(fp,"#### ID:RANK-MP-CASE\n#### Class k~, Case ID~, ");
  if (mode_eval) conv_fprintf(fp,"Answer class~, ");
  if (tuple_size > 2) {
    conv_fprintf(fp,"Attr. x1~, ...~, Attr. x%d~, ",tuple_size);
  }
  else {
    conv_fprintf(fp,"Attr. x1~, Attr. x2~, ");
  }
  conv_fprintf(fp,"Class prob. P(k)~, Membership prob. P(k|x)\n"
	  "#### (in the decending order of the class probs. and"
	  " the membership probs.)\n####\n");
  for (k=0; k<num_class; k++) {
    k0 = rank_class[k];
    for (e=0; e<num_show_case_rank; e++) {
      j = rank_case_mem[k0][e];
      conv_fprintf(fp,"%d,%d,",k0,j);
      if (mode_eval) conv_fprintf(fp,"%s,",ans_class_list[freq_list[j].ans_id]);
      for (v=0; v<tuple_size; v++)
	conv_fprintf(fp,"%s,",word_list[v][freq_list[j].id[v]]);
      conv_fprintf(fp,"%.12Lf,%.12Lf\n",
	      param_c[k0],prob_case_mem[k0][j]);
    }
  }

  for (v=0; v<tuple_size; v++) {
    conv_fprintf(fp,"#### ID:RANK-MP-ATT%d\n#### Class k~, Attr. value x%d~, "
		 "Class prob. P(k)~, Membership prob. P(k|x%d)\n"
		 "#### (in the decending order of the class probs. and"
		 " the membership probs.)\n####\n",v,v,v);
    for (k=0; k<num_class; k++) {
      k0 = rank_class[k];
      for (i=0; i<num_show_att_rank[v]; i++) {
	j = rank_att_mem[k0][v][i];
	conv_fprintf(fp,"%d,%s,%.12Lf,%.12Lf\n",
		k0,word_list[v][j],param_c[k0],prob_att_mem[k0][v][j]);
      }
    }
  }

  free(num_show_att_rank);

  fflush(fp);
}

void print_rank_wodd(FILE *fp)
{
  int k,k0,i,j,v,e;
  int num_show_case_rank;
  int *num_show_att_rank;

  num_show_att_rank = (int *)malloc(sizeof(int)*tuple_size);
  if (num_show_att_rank == NULL)
    die("Couldn't allocate memory for num_show_att_rank\n");

  num_show_case_rank = (num_rank > 0 && num_rank < num_entry) ? num_rank : num_entry;
  for (v=0; v<tuple_size; v++)
    num_show_att_rank[v] = (num_rank > 0 && num_rank < num_word[v]) ? num_rank : num_word[v];

  conv_fprintf(fp,"#### ID:RANK-WODD-CASE\n#### Class k~, Case ID~, ");
  if (mode_eval) conv_fprintf(fp,"Answer class~, ");
  if (tuple_size > 2) {
    conv_fprintf(fp,"Attr. x1~, ...~, Attr. x%d~, ",tuple_size);
  }
  else {
    conv_fprintf(fp,"Attr. x1~, Attr. x2~, ");
  }
  conv_fprintf(fp,"Class prob. P(k)~, Membership prob. P(k|x)~, "
	       "WLLR P(x|k)log{P(x|k)/P(x|not k)}\n"
	       "#### (in the decending order of the class probs. and"
	       " WLLR)\n####\n");
  for (k=0; k<num_class; k++) {
    k0 = rank_class[k];
    for (e=0; e<num_show_case_rank; e++) {
      j = rank_case_mem[k0][e];
      conv_fprintf(fp,"%d,%d,",k0,j);
      if (mode_eval) conv_fprintf(fp,"%s,",ans_class_list[freq_list[j].ans_id]);
      for (v=0; v<tuple_size; v++)
	conv_fprintf(fp,"%s,",word_list[v][freq_list[j].id[v]]);
      conv_fprintf(fp,"%.12Lf,%.12Lf,%.12Lf\n",
	      param_c[k0],prob_case_mem[k0][j],wodd_case[k0][j]);
    }
  }

  for (v=0; v<tuple_size; v++) {
    conv_fprintf(fp,"#### ID:RANK-WODD-ATT%d\n#### Class k~, Attr. x%d~, "
		 "Class prob. P(k)~, Membership prob. P(k|x%d)~, "
		 "WLLR P(x%d|k)log{P(x%d|k)/P(x%d|not k)}\n"
		 "#### (in the descending order of the class probs. and"
		 " WLLRs)\n####\n",v,v,v,v,v,v);
    for (k=0; k<num_class; k++) {
      k0 = rank_class[k];
      for (i=0; i<num_show_att_rank[v]; i++) {
	j = rank_att_mem[k0][v][i];
	conv_fprintf(fp,"%d,%s,%.12Lf,%.12Lf,%.12Lf\n",
		k0,word_list[v][j],param_c[k0],prob_att_mem[k0][v][j],wodd_att[k0][v][j]);
      }
    }
  }

  free(num_show_att_rank);

  fflush(fp);
}

void print_rank(FILE *fp)
{
  switch (rel_m) {
  case REL_MP:
    print_rank_ml(fp);
    break;
  case REL_WODD:
    print_rank_wodd(fp);
    break;
  default:
    fprintf(stderr,"Internal error: Unexpected relevance measure between class and attribute values\n");
    exit(1);
  }
}

void print_rank_sm(FILE *fp,int t,int *cand)
{
  int k0,k1,k2,k3,i,j,l,v;
  int *num_show_att_rank;

  num_show_att_rank = (int *)malloc(sizeof(int)*tuple_size);
  if (num_show_att_rank == NULL)
    die("Couldn't allocate memory for num_show_att_rank\n");

  for (v=0; v<tuple_size; v++)
    num_show_att_rank[v] =
      (num_rank_sm > 0 && num_rank_sm < num_word[v]) ? num_rank_sm : num_word[v];

  k1 = cand[0];
  k2 = cand[1];
  k3 = cand[2];

  for (v=0; v<tuple_size; v++) {

    conv_fprintf(fp,"#### ID:RANK-SMEM-BEFORE-ATT%d\n",v);
    conv_fprintf(fp,"#### Class k~, Attr. value x%d~, Class prob. P(k)~, Membership prob. P(k|x%d)\n#### (in the descending order of membership probs.)\n",v,v);
    conv_fprintf(fp,"#### Split-merge operation %d: [%d~,%d] -> [%d]~, [%d] -> [%d~,%d]\n####\n",
		 t,k1,k2,k1,k3,k2,k3);
    
    for (l=0; l<3; l++) {
      k0 = cand[l];
      for (i=0; i<num_show_att_rank[v]; i++) {
	j = rank_att_mem_sm[k0][v][i];
	conv_fprintf(fp,"%d,%s,%.12Lf,%.12Lf\n",
		     k0,word_list[v][j],param_c_sm[k0],prob_att_mem_sm[k0][v][j]);
      }
    }

    conv_fprintf(fp,"#### ID:RANK-SMEM-AFTER-ATT%d\n",v);
    conv_fprintf(fp,"#### Class k~, Attr. value x%d~, Class prob. P(k)~, Membership prob. P(k|x%d)\n#### (in the descending order of membership probs.)\n",v,v);
    conv_fprintf(fp,"#### Split-merge operation %d: [%d~,%d] -> [%d]~, [%d] -> [%d~,%d]\n####\n",
		 t,k1,k2,k1,k3,k2,k3);

    for (l=0; l<3; l++) {
      k0 = cand[l];
      for (i=0; i<num_show_att_rank[v]; i++) {
	j = rank_att_mem_part[l][v][i];
	conv_fprintf(fp,"%d,%s,%.12Lf,%.12Lf\n",
		k0,word_list[v][j],param_c[k0],prob_att_mem_part[l][v][j]);
      }
    }
  }

  free(num_show_att_rank); num_show_att_rank = NULL;

  fflush(fp);
}

void print_case_kl_dist_header(FILE *fp)
{
  if (!mode_eval)
    conv_fprintf(fp,"#### ID:DISSIM-KL-CASE\n#### Case ID of x~, Case ID of x'~, "
		 "KL-divergence D(x||x')\n####\n");
  else
    conv_fprintf(fp,"#### ID:DISSIM-KL-CASE\n#### Case ID of x~, Answer class of x~, "
		 "Case ID of x'~, Answer class of x'~, "
		 "KL-divergence D(x||x')\n####\n");
}

void print_att_kl_dist_header(FILE *fp, int v)
{
  conv_fprintf(fp,"#### ID:DISSIM-KL-ATT%d\n#### Attr. value x%d~, Attr. value x%d'~, "
	       "KL-divergence D(x%d||x%d')\n####\n",v,v,v,v,v);
}

void print_case_kl_dist(FILE *fp, int e0, int e1, FLOAT_TYPE d)
{
  if (!mode_eval)
    conv_fprintf(fp,"%d,%d,%.12Lf\n",e0,e1,d);
  else
    conv_fprintf(fp,"%d,%s,%d,%s,%.12Lf\n",
	    e0,ans_class_list[freq_list[e0].ans_id],
	    e1,ans_class_list[freq_list[e1].ans_id],
	    d);
}

void print_att_kl_dist(FILE *fp, int v, int i, int j, FLOAT_TYPE d)
{
  conv_fprintf(fp,"%s,%s,%.12Lf\n",word_list[v][i],word_list[v][j],d);
}

#endif
